#include <stdio.h>
#include <string.h>
#include <inttypes.h>
#include <cassert>
using namespace std;

#include "cbp3_def.h"
#include "cbp3_framework.h"

#include "predictor.h"

/*********************************************
 * FTL++
 *********************************************/
//history length for global components
int lhl[] = { 5, 9, 16, 29, 52, 90, 163, 292 };
//history length for per-set32 components
int ps32hl[] = { 6, 11, 18, 28, 45, 71 };
//history length for per-set8 components
int ps8hl[] = { 7, 15 };
class FTLPP {
protected:
	static const int TABBITS = 12;
	static const int BIMBITS = TABBITS;//number of entries of BIM

	//counter width of pattern history table
	static const int BBIT = 6;//counter width of BIM
	static const int CBIT = 6;//counter width of PHT(Global,Per-Set)
	static const int LBIT = 6;//counter width of PHT(LOCAL)
	static const int GBIT = 6;//counter width of PHT(GAg)
	//width of threshold counter
	static const int BCBIT = 10; //BC width
	static const int TCBIT = CBIT + 6;//TC width

	//Local History Table
	static const int LHTBITS = 10;
	static const int LHTWIDTH = 3;//width of local history table
	static const uint32_t LHTSIZE = 1 << LHTBITS;//number of entries of local history table

	//BIM
	static const uint32_t BIMSIZE = 1 << BIMBITS;
	//PHT
	static const uint32_t TABSIZE = 1 << TABBITS;
	static const uint32_t LHSIZE = 1 << BIMBITS;

	//Per-Set
	static const int PS8 = 3;
	static const int PS32 = 5;
	static const int PS8LINES = 1 << PS8;//num of entries of per-set8 (=8)
	static const int PS32LINES = 1 << PS32;//num of entries of per-set32 (=32)

	//num of PHT
	static const int GLOBAL = 8; //Global : 8 components
	static const int PS8NUM = 2; //Per-Set8 : 2 components
	static const int PS32NUM = 6; //Per-Set32 : 6 components
	//GAG : 1 components
	//Local : 1components

	//threshold for exception filter and initial value of TC
	static const int T = GLOBAL + PS8NUM + PS32NUM + 3;

	//circular buffer size
	static const int BSIZE = 128;

	//Kernel-User
	static const int SPACE = 2;
	static const int KERNEL = 1;
	static const int USER = 0;

	//place of pc for history register with nonConditional branch
	static const int NCONDHS = 2;
protected:
	//Local History Table
	uint32_t LHT[1 << LHTBITS]; //retire
	uint32_t LHT_f[1 << LHTBITS]; //fetch

	//history registers
	HISTORY ghr[SPACE];//retire
	HISTORY phr[SPACE];//retire
	HISTORY ghr_f[SPACE];//fetch
	HISTORY phr_f[SPACE];//fetch

	//Per-Set History Tables
	//retire
	HISTORY ps8hr[PS8LINES];
	HISTORY ps8phr[PS8LINES];
	//fetch
	HISTORY ps8hr_f[PS8LINES];
	HISTORY ps8phr_f[PS8LINES];
	//retire
	HISTORY ps32hr[PS32LINES];
	HISTORY ps32phr[PS32LINES];
	//fetch
	HISTORY ps32hr_f[PS32LINES];
	HISTORY ps32phr_f[PS32LINES];

	//folded histories
	//global
	FOLDHIST<TABBITS> gidx[SPACE][GLOBAL];//retire
	FOLDHIST<TABBITS> gidx_f[SPACE][GLOBAL];//fetch
	//per-set8
	FOLDHIST<TABBITS> ps8idx[PS8NUM][PS8LINES];//retire
	FOLDHIST<TABBITS> ps8idx_f[PS8NUM][PS8LINES];//fetch
	//per-set32
	FOLDHIST<TABBITS> ps32idx[PS32NUM][PS32LINES];//retire
	FOLDHIST<TABBITS> ps32idx_f[PS32NUM][PS32LINES];//fetch

	//threshold counter
	UCOUNTER<TCBIT> TC;
	//bim counter
	SCOUNTER<BCBIT> BC;

	//----------Pattern History Tables----------
	//BIM : 1 * 2^12 * 6 = 24,576
	SCOUNTER<BBIT> bim[BIMSIZE];
	//GAg : 1 * 2^12 * 6 = 24,576
	SCOUNTER<GBIT> gpht[LHSIZE];
	//Local : 1 * 2^12 * 6 = 24,576
	SCOUNTER<LBIT> lhpht[LHSIZE];
	//others : (2+6+8) * 2^12 * 6 = 393,216
	SCOUNTER<CBIT> pht[GLOBAL + PS8NUM + PS32NUM][TABSIZE];

	//-----------White list filters--------------
	//Bias Filter
	biaspredictor<14, true> bfilter;
	//Loop Filter
	looppredictor<16, 8, 15, true> lfilter;

	//-----------Black list filters--------------
	//Exceptional Filter
	exceptionalfilter<8, 8, 15, true> exfilter;
	//Give up Filter
	giveupfilter<8, 8, 15, true> gufilter;

	//buffer
	uint32_t BUSEbuf[BSIZE];
	uint32_t TCbuf[BSIZE];
	uint32_t pbuf[BSIZE];
	//buffer pointer
	uint32_t pptr;
	uint32_t rptr;

protected:

	/////////////// Hash Functions ////////////////
	// index of BIM
	uint32_t bidx(uint32_t pc) {
		return ((pc ^ (pc >> (BIMBITS))) & (BIMSIZE - 1));
	}

	//index of the table that directly indexed by Global history
	uint32_t gagidx_fetch(uint32_t pc) {//fetch stage
		int space = ((pc & 0xc0000000) == 0xc0000000) ? 1 : 0;
		return (ghr_f[space].read(0, TABBITS));
	}
	uint32_t gagidx(uint32_t pc) {//retire stage
		int space = ((pc & 0xc0000000) == 0xc0000000) ? 1 : 0;
		return (ghr[space].read(0, TABBITS));
	}

	//index of Local history table
	uint32_t lhtidx(uint32_t pc) {
		return ((pc) ^ (pc >> (32 - LHTBITS))) & ((1 << LHTBITS) - 1);
	}
	//index of PHT(Local)
	uint32_t lhidx_fetch(uint32_t pc) {//fetch stage
		return (((pc ^ (pc >> (TABBITS - 1))) ^ ((pc >> (TABBITS * 2 - 1) & 1))) ^ ((LHT_f[lhtidx(pc)] & ((1
				<< LHTWIDTH) - 1)) << (TABBITS - LHTWIDTH))) & (TABSIZE - 1);
	}
	uint32_t lhidx(uint32_t pc) {//retire stage
		return (((pc ^ (pc >> (TABBITS - 1))) ^ ((pc >> (TABBITS * 2 - 1) & 1))) ^ ((LHT[lhtidx(pc)] & ((1 << LHTWIDTH)
				- 1)) << (TABBITS - LHTWIDTH))) & (TABSIZE - 1);
	}

	///////////////////// Prediction //////////////////////
	// Prediction values in fetch stage
	int getsum_fetch(uint32_t pc) {
		int space = ((pc & 0xc0000000) == 0xc0000000) ? 1 : 0;//detect space

		int sum = 0;

		//Sum up counter values in PHTs

		//BIM
		sum += bim[bidx(pc)].read();

		//Global
		for (int i = 0; i < GLOBAL; i++) {
			sum += pht[i][gidx_f[space][i].read(pc, i)].read();
		}

		//Per-Set8
		int set;
		set = pc & ((1 << PS8) - 1);// detect set
		if (space == KERNEL) //Limit Set if in the kernel space
			set %= 4;
		for (int i = 0; i < PS8NUM; i++) {
			sum += pht[i + GLOBAL][ps8idx_f[i][set].read(pc >> PS8, i)].read();
		}

		//Per-Set32
		set = pc & ((1 << PS32) - 1); //detect set
		if (space == KERNEL) //Limit Set if in the kernel space
			set %= 4;
		for (int i = 0; i < PS32NUM; i++) {
			sum += pht[i + GLOBAL + PS8NUM][ps32idx_f[i][set].read(pc >> PS32, i + PS8NUM)].read();
		}

		//GAg
		sum += gpht[gagidx_fetch(pc)].read();
		//Local
		sum += lhpht[lhidx_fetch(pc)].read();

		return sum;
	}
	// Prediction values in retire stage
	int getsum(uint32_t pc) {
		int space = ((pc & 0xc0000000) == 0xc0000000) ? 1 : 0;//detect space
		int sum = 0;

		//Sum up counter values in phts
		//BIM
		sum += bim[bidx(pc)].read();

		//Global
		for (int i = 0; i < GLOBAL; i++) {
			sum += pht[i][gidx[space][i].read(pc, i)].read();
		}

		//Per-Set8
		int set;
		set = pc & ((1 << PS8) - 1);
		if (space == KERNEL)
			set %= 4;
		for (int i = 0; i < PS8NUM; i++) {
			sum += pht[i + GLOBAL][ps8idx[i][set].read(pc >> PS8, i)].read();
		}

		//Per-Set32
		set = pc & ((1 << PS32) - 1);
		if (space == KERNEL)
			set %= 4;
		for (int i = 0; i < PS32NUM; i++) {
			sum += pht[i + GLOBAL + PS8NUM][ps32idx[i][set].read(pc >> PS32, i + PS8NUM)].read();
		}

		//GAg
		sum += gpht[gagidx(pc)].read();
		//Local
		sum += lhpht[lhidx(pc)].read();

		return sum;
	}

	/////////////////// Update Patturn History Tables //////////////////////
	//update PHTs in retire stage
	void updateTables(uint32_t pc, bool taken, bool pred) {
		int space = ((pc & 0xc0000000) == 0xc0000000) ? 1 : 0;//detect space

		//BIM
		bim[bidx(pc)].update(taken);

		//Global
		for (int i = 0; i < GLOBAL; i++) {
			pht[i][gidx[space][i].read(pc, i)].update(taken);
		}

		//Per-Set8
		int set;
		set = pc & ((1 << PS8) - 1);
		if (space == KERNEL)
			set %= 4;
		for (int i = 0; i < PS8NUM; i++) {
			pht[i + GLOBAL][ps8idx[i][set].read(pc >> PS8, i)].update(taken);
		}

		//Per-Set32
		set = pc & ((1 << PS32) - 1);
		if (space == KERNEL)
			set %= 4;
		for (int i = 0; i < PS32NUM; i++) {
			pht[i + GLOBAL + PS8NUM][ps32idx[i][set].read(pc >> PS32, i + PS8NUM)].update(taken);
		}
		//GAg
		gpht[gagidx(pc)].update(taken);
		//Local
		lhpht[lhidx(pc)].update(taken);
	}

	/////////////////// History housekeeping //////////////////////
	// History housekeeping in fetch stage
	void updateHist_fetch(uint32_t pc, bool taken, bool cond) {
		int space = ((pc & 0xc0000000) == 0xc0000000) ? 1 : 0;//detect space

		//if conditional branch comes, update local history table
		if (cond) {
			LHT_f[lhtidx(pc)] = (((LHT_f[lhtidx(pc)] << 1) | (taken ? 1 : 0))) & ((1 << LHTWIDTH) - 1);
		}

		//Per-Set8
		int set = pc & ((1 << PS8) - 1);
		if (space == KERNEL)
			set %= 4;

		ps8phr_f[set].push((pc & (1 << PS8)) != 0);
		if (!cond) {
			ps8hr_f[set].push((pc & (1 << (PS8 + NCONDHS))) != 0);
		} else {
			ps8hr_f[set].push(taken);
		}

		for (int i = 0; i < PS8NUM; i++) {
			ps8idx_f[i][set].update(ps8hr_f[set], ps8phr_f[set]);
		}

		//Per-Set32
		set = pc & ((1 << PS32) - 1);
		if (space == KERNEL)
			set %= 4;
		ps32phr_f[set].push((pc & (1 << PS32)) != 0);
		if (!cond) {
			ps32hr_f[set].push((pc & (1 << (PS32 + NCONDHS))) != 0);
		} else {
			ps32hr_f[set].push(taken);
		}
		for (int i = 0; i < PS32NUM; i++) {
			ps32idx_f[i][set].update(ps32hr_f[set], ps32phr_f[set]);
		}

		//Global
		//update user history if in user space
		if (space == USER) {
			if (!cond) {
				ghr_f[USER].push((pc & (1 << NCONDHS)) != 0);
			} else {
				ghr_f[USER].push(taken);
			}
			phr_f[USER].push((pc & (1 << 0)) != 0);

			for (int i = 0; i < GLOBAL; i++) {
				gidx_f[USER][i].update(ghr_f[USER], phr_f[USER]);
			}
		}

		//always update kernel history
		if (!cond) {
			ghr_f[KERNEL].push((pc & (1 << NCONDHS)) != 0);
		} else {
			ghr_f[KERNEL].push(taken);
		}
		phr_f[KERNEL].push((pc & (1 << 0)) != 0);

		for (int i = 0; i < GLOBAL; i++) {
			gidx_f[KERNEL][i].update(ghr_f[KERNEL], phr_f[KERNEL]);
		}
	}
	// History housekeeping in retire stage
	void updateHist(uint32_t pc, bool taken, bool cond) {
		int space = ((pc & 0xc0000000) == 0xc0000000) ? 1 : 0;//detecting space

		//if conditional branch comes, update local history table
		if (cond) {
			LHT[lhtidx(pc)] = (((LHT[lhtidx(pc)] << 1) | (taken ? 1 : 0))) & ((1 << LHTWIDTH) - 1);
		}

		//Per-Set8
		int set;
		set = pc & ((1 << PS8) - 1);
		if (space == KERNEL) {
			set %= 4;
		}
		ps8phr[set].push((pc & (1 << PS8)) != 0);
		if (!cond) {
			ps8hr[set].push((pc & (1 << (PS8 + NCONDHS))) != 0);
		} else {
			ps8hr[set].push(taken);
		}

		for (int i = 0; i < PS8NUM; i++) {
			ps8idx[i][set].update(ps8hr[set], ps8phr[set]);
		}

		//Per-Set32
		set = pc & ((1 << PS32) - 1);
		if (space == KERNEL) {
			set %= 4;
		}
		ps32phr[set].push((pc & (1 << PS32)) != 0);
		if (!cond) {
			ps32hr[set].push((pc & (1 << (PS32 + NCONDHS))) != 0);
		} else {
			ps32hr[set].push(taken);
		}
		for (int i = 0; i < PS32NUM; i++) {
			ps32idx[i][set].update(ps32hr[set], ps32phr[set]);
		}

		//Global
		//update user history if in user space
		if (space == USER) {
			if (!cond) {
				ghr[USER].push((pc & (1 << NCONDHS)) != 0);
			} else {
				ghr[USER].push(taken);
			}
			phr[USER].push((pc & (1 << 0)) != 0);
			for (int i = 0; i < GLOBAL; i++) {
				gidx[USER][i].update(ghr[USER], phr[USER]);
			}
		}

		//always update kernel history
		if (!cond) {
			ghr[KERNEL].push((pc & (1 << NCONDHS)) != 0);
		} else {
			ghr[KERNEL].push(taken);
		}
		phr[KERNEL].push((pc & (1 << 0)) != 0);
		for (int i = 0; i < GLOBAL; i++) {
			gidx[KERNEL][i].update(ghr[KERNEL], phr[KERNEL]);
		}

	}

	// Initialize Folded Histories
	void initHist() {
		//Global
		for (int i = 0; i < GLOBAL; i++) {
			for (int j = 0; j < SPACE; ++j) {
				gidx[j][i].init(lhl[i]);
				gidx_f[j][i].init(lhl[i]);
			}
		}
		//Per-Set8
		for (int i = 0; i < PS8NUM; i++) {
			for (int s = 0; s < PS8LINES; s++) {
				ps8idx[i][s].init(ps8hl[i]);
				ps8idx_f[i][s].init(ps8hl[i]);
			}
		}
		//Per-Set32
		for (int i = 0; i < PS32NUM; i++) {
			for (int s = 0; s < PS32LINES; s++) {
				ps32idx[i][s].init(ps32hl[i]);
				ps32idx_f[i][s].init(ps32hl[i]);
			}
		}
	}

public:

	//budget counting
	int budgets() {
		printf("budget counting:\n");
		int sum = 0, b;

		//-------------PHT-------------
		b = BIMSIZE * CBIT;
		sum += b;
		printf("BIM:\t%d\n", b);

		b = TABSIZE * CBIT * GLOBAL;
		sum += b;
		printf("Global PHT:\t%d\n", b);

		b = TABSIZE * CBIT * PS32NUM;
		sum += b;
		printf("Per-Set(32) PHT:\t%d\n", b);

		b = TABSIZE * CBIT * PS8NUM;
		sum += b;
		printf("Per-Set(8) PHT:\t%d\n", b);

		b = TABSIZE * CBIT;
		sum += b;
		printf("GAg PHT:\t%d\n", b);

		b = TABSIZE * CBIT;
		sum += b;
		printf("Local PHT:\t%d\n", b);

		//------HISTORY TABLE/HISTORY REGISTER--------
		b = (lhl[GLOBAL - 1]+1) * 2 * 2;
		sum += b;
		printf("GHR:\t%d\n", b);

		b = (32+1) * 2 * 2;
		sum += b;
		printf("PHR:\t%d\n", b);

		b = (ps32hl[PS32NUM - 1]+1) * PS32LINES * 2;
		sum += b;
		printf("Per-Set(32)HT:\t%d\n", b);

		b = (32+1) * PS32LINES * 2;
		sum += b;
		printf("Per-Set(32)PHR:\t%d\n", b);

		b = (ps8hl[PS8NUM - 1]+1) * PS8LINES * 2 ;
		sum += b;
		printf("Per-Set(8)HT:\t%d\n", b);

		b = (ps8hl[PS8NUM - 1]+1) * PS8LINES * 2 ;
		sum += b;
		printf("Per-Set(8)PHR:\t%d\n", b);

		b = (LHTWIDTH) * (1 << LHTBITS) * 2;
		sum += b;
		printf("LHT:\t%d\n", b);

		//----------FOLDED HISTORY-----------
		b = GLOBAL * TABBITS * 2 * 2;
		sum += b;
		printf("global folded history:\t%d\n", b);

		b = PS32NUM * TABBITS * PS32LINES * 2;
		sum += b;
		printf("Per-Set(32)folded history:\t%d\n", b);

		b = PS8NUM * TABBITS * PS8LINES * 2;
		sum += b;
		printf("Per-Set(8) folded history:\t%d\n", b);

		//-----------FILTER------------
		b = bfilter.budgets();
		sum += b;
		printf("bias filter:\t%d\n", b);

		b = lfilter.budgets();
		sum += b;
		printf("loop filter:\t%d\n", b);

		b = exfilter.budgets();
		sum += b;
		printf("exception filter:\t%d\n", b);

		b = gufilter.budgets();
		sum += b;
		printf("give up filter:\t%d\n", b);

		//-----------OTHERS------------
		b = BSIZE * (1 + 1 + 1);
		sum += b;
		printf("Circular Buffer:\t%d\n", b);

		b = 7 + 7;
		sum += b;
		printf("Buffer pointers:\t%d\n", b);

		b = TCBIT;
		sum += b;
		printf("TC:\t%d\n", b);

		b = BCBIT;
		sum += b;
		printf("BC:\t%d\n", b);

		b = 2;
		sum += b;
		printf("call/return flag:\t%d\n", b);

		b = 32 * 2;
		sum += b;
		printf("pc buffer:\t%d\n", b);

		return sum;
	}
	void end() {
		printf("Total budget count:\t%d\n", budgets()); // checks total budget is less than 532,480 bits (65KB)
	}
	//Initialize
	void init() {
		memset(bim, 0, sizeof(bim));
		memset(gpht, 0, sizeof(gpht));
		memset(lhpht, 0, sizeof(lhpht));
		memset(pht, 0, sizeof(pht));

		for (int j = 0; j < SPACE; ++j) {
			ghr[j].init(lhl[GLOBAL - 1]);
			phr[j].init(32);
			ghr_f[j].init(lhl[GLOBAL - 1]);
			phr_f[j].init(32);
		}
		for (int i = 0; i < PS8LINES; ++i) {
			ps8hr[i].init(ps8hl[PS8NUM - 1]);
			ps8phr[i].init(ps8hl[PS8NUM - 1]);
			ps8hr_f[i].init(ps8hl[PS8NUM - 1]);
			ps8phr_f[i].init(ps8hl[PS8NUM - 1]);
		}
		for (int i = 0; i < PS32LINES; ++i) {
			ps32hr[i].init(ps32hl[PS32NUM - 1]);
			ps32phr[i].init(32);
			ps32hr_f[i].init(ps32hl[PS32NUM - 1]);
			ps32phr_f[i].init(32);
		}

		for (int i = 0; i < 1 << LHTBITS; ++i) {
			LHT[i] = 0;
			LHT_f[i] = 0;
		}
		initHist();

		TC.write((1 << 7) * T);
		BC.write(0);
		pptr = rptr = 0;
	}

	uint32_t predict(uint32_t pc, bool cond, bool indir) {
		if (!cond) {
			return 0;
		}
		int sum = getsum_fetch(pc);//get sum value of tables

		//BIM use flag
		BUSEbuf[pptr] = abs(sum) >= (TC.read() >> 7);
		//Prediction
		pbuf[pptr] = (sum >= 0);
		//TC
		TCbuf[pptr] = ((abs(sum) < (TC.read() >> 6)));
		//update pointer
		pptr = (pptr + 1) % BSIZE;

		bool filtered = false;

		//--------Filtering----------
		filtered = bfilter.hit(pc);
		if (filtered)//Bias filtering
			return bfilter.pred(pc);

		filtered = lfilter.filterhit(pc);
		if (filtered) {//Loop filtering
			return lfilter.pred(pc);
		}

		filtered = gufilter.filterhit(pc);
		if (filtered)//Give up filtering
			return gufilter.pred(pc);

		bool pred = false;
		//BIM over writing
		if ((BC.read() >= 0) && (abs(sum) < (TC.read() >> 7))) {
			//Use BIM for prediction
			pred = (bim[bidx(pc)].read() >= 0) ? 1 : 0;
		} else {
			//use prediction of base predictor
			pred = (sum >= 0) ? 1 : 0;
		}
		return pred;
	}

	//update in fetch stage
	void update_fetch(uint32_t pc, bool cond, bool indir, bool taken) {
		updateHist_fetch(pc, taken, cond);
		if (cond) {
			//update loop filter
			lfilter.update_fetch(pc, taken);
		}
	}

	//update in retire stage (Full update)
	void update(uint32_t pc, bool cond, bool indir, bool taken) {
		if (!cond) {
			//Update History
			updateHist(pc, taken, cond);
			return;
		}

		//pop information from circular Buffer
		bool pred = pbuf[rptr];//prediction in fetch stage
		bool tc_b = TCbuf[rptr];//abs(sum) < TC
		bool buse = BUSEbuf[rptr];//BC overwrote
		rptr = (rptr + 1) % BSIZE;

		//--------------Filtering-------------
		bool do_update = true; //if update is filtered, do_update becomes false
		do_update = !bfilter.update(pc, taken); //bias filtering
		if (do_update) {
			lfilter.update(pc, taken);
			do_update = !lfilter.filterhit(pc); //loop filtering
		}
		if (do_update) {
			gufilter.update(pc, taken, pred);
			do_update = !gufilter.hpred(pc); //give up filtering
		}

		//Update Base predictor
		if (do_update) {
			int sum = getsum(pc);// sum in retire

			//update exceptional filter
			if ((pred != taken) && (abs(sum) > ((T * 3)))) {
				//not worth training case candidate
				exfilter.update(pc, (pred == taken), true);
			} else {
				exfilter.update(pc, (pred == taken), false);
			}

			//update BC
			if (buse && (pred != (bim[bidx(pc)].read() >= 0))) {
				BC.update((bim[bidx(pc)].read() >= 0) == taken);
			}

			//exceptional filtering
			if ((pred != taken) && (abs(sum) > ((T * 3))) && exfilter.filterhit(pc)) {
				//not worth training case
				//Don't update base predictor
			} else if ((pred != (sum >= 0)) || tc_b || (pred != taken)) {
				//update PHT
				updateTables(pc, taken, pred);
				//update TC
				if (pred != (sum >= 0)) {
					TC.update(true);
				}
				TC.update(pred != taken);
			} else if ((bim[bidx(pc)].read() * (taken ? 1 : -1)) < 2) {
				//accelerate training
				bim[bidx(pc)].update(taken);
			}
		}
		//Update History
		updateHist(pc, taken, cond);
	}
};


// predictor
FTLPP *predictor;

//call/return flag
bool fnext = false, next = false;
//pc buffer
uint32_t pcf, pc;


void PredictorInit() {
}

void PredictorReset() {
	fnext = false, next = false;
	pcf = pc = 0;
	predictor = new FTLPP();
	predictor->init();
}

void PredictorRunACycle() {
	// get info about what uops are processed at each pipeline stage
	const cbp3_cycle_activity_t *cycle_info = get_cycle_info();
	// make prediction at fetch stage
	for (int i = 0; i < cycle_info->num_fetch; i++) {
		uint32_t fe_ptr = cycle_info->fetch_q[i];
		const cbp3_uop_dynamic_t *uop = &fetch_entry(fe_ptr)->uop;
		// Decode instruction
		bool call = (uop->type & IS_BR_CALL) != 0;//call
		bool ret = (uop->type & IS_BR_RETURN) != 0;//return
		bool other = (uop->type & IS_BR_OTHER) != 0;//other
		bool cond = (uop->type & IS_BR_CONDITIONAL) != 0;//conditional
		bool indir = (uop->type & IS_BR_INDIRECT) != 0;//indirect

		bool taken = (uop->br_taken) || (!cond);
		uint32_t gpred = 0;

		if (fnext) {//push call/return target address to history
			fnext = false;
			predictor->update_fetch(uop->pc, false, false, true);
		} else if ((((pcf & 0xc0000000) != 0xc0000000) && ((uop->pc & 0xc0000000) == 0xc0000000))) {
			//if jump to kernel space,push pc to history
			predictor->update_fetch(uop->pc, false, false, true);
		}

		// Check whether the target instruction is brach.
		if (uop_is_branch(uop->type)) {
			// Prediction Phase
			gpred = predictor->predict(uop->pc, cond, indir);
			if (cond) {
				assert(report_pred(fe_ptr, false, gpred));
			}
			// Update phase
			predictor->update_fetch(uop->pc, cond, indir, taken);

			if (call || other) {//if call/other, push history twice
				predictor->update_fetch(uop->pc, false, false, true);
			}
		}
		if (call || ret) {
			fnext = true;//remebmer call,return
		}
		pcf = uop->pc;
	}

	for (int i = 0; i < cycle_info->num_retire; i++) {
		uint32_t rob_ptr = cycle_info->retire_q[i];
		const cbp3_uop_dynamic_t *uop = &rob_entry(rob_ptr)->uop;

		bool call = (uop->type & IS_BR_CALL) != 0;//call
		bool ret = (uop->type & IS_BR_RETURN) != 0;//return
		bool other = (uop->type & IS_BR_OTHER) != 0;//other
		bool cond = (uop->type & IS_BR_CONDITIONAL) != 0;//conditional
		bool indir = (uop->type & IS_BR_INDIRECT) != 0;//indirect

		bool taken = (uop->br_taken) || (!cond);

		if (next) {//push call/return target address to history
			next = false;
			predictor->update(uop->pc, false, false, true);
		} else if ((((pc & 0xc0000000) != 0xc0000000) && ((uop->pc & 0xc0000000) == 0xc0000000))) {
			//if jump to kernel space,push pc to history
			predictor->update(uop->pc, false, false, true);
		}

		if (uop_is_branch(uop->type)) {
			predictor->update(uop->pc, cond, indir, taken);
			if (call || other) {//if call/other, push history twice
				predictor->update(uop->pc, false, false, true);
			}
		}
		if (call || ret) {
			next = true;//remebmer call,rreturn
		}
		pc = uop->pc;
	}

}

void PredictorRunEnd() {
	predictor->end();
	delete predictor;
	rewind_marked = false;
}

void PredictorExit() {
}
